@@ -1,5 +1,5 @@
 /******************************************************************
- * CopyPolicy: GNU Lesser General Public License 3 applies
+ * CopyPolicy: GNU Lesser General Public License 2.1 applies
  * Copyright (C) 1998-2008 Monty xiphmont@mit.edu
  * 
  * Autoscan for or verify presence of a cdrom device
@@ -14,6 +14,7 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include <pwd.h>
+#include <time.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include "cdda_interface.h"
@@ -263,8 +264,13 @@ cdrom_drive *cdda_identify_cooked(const char *dev, int messagedest,
   d->interface=COOKED_IOCTL;
   d->bigendianp=-1; /* We don't know yet... */
   d->nsectors=-1;
+  d->private=calloc(1,sizeof(*d->private));
+  {
+    /* goddamnit */
+    struct timespec tv;
+    d->private->clock=(clock_gettime(CLOCK_MONOTONIC,&tv)<0?CLOCK_REALTIME:CLOCK_MONOTONIC);
+  }
   idmessage(messagedest,messages,"\t\tCDROM sensed: %s\n",description);
-  
   return(d);
 }
 
@@ -668,10 +674,15 @@ cdrom_drive *cdda_identify_scsi(const char *generic_device,
   d->bigendianp=-1; /* We don't know yet... */
   d->nsectors=-1;
   d->messagedest = messagedest;
-
+  d->private=calloc(1,sizeof(*d->private));
+  {
+    /* goddamnit */
+    struct timespec tv;
+    d->private->clock=(clock_gettime(CLOCK_MONOTONIC,&tv)<0?CLOCK_REALTIME:CLOCK_MONOTONIC);
+  }
   if(use_sgio){
     d->interface=SGIO_SCSI;
-    d->sg_buffer=d->sg=malloc(MAX_BIG_BUFF_SIZE);
+    d->private->sg_buffer=(unsigned char *)(d->private->sg_hd=malloc(MAX_BIG_BUFF_SIZE));
     g_fd=d->cdda_fd=dup(d->ioctl_fd);
   }else{
     version=verify_SG_version(d,messagedest,messages);
@@ -685,8 +696,8 @@ cdrom_drive *cdda_identify_scsi(const char *generic_device,
     }
 
     /* malloc our big buffer for scsi commands */
-    d->sg=malloc(MAX_BIG_BUFF_SIZE);
-    d->sg_buffer=d->sg+SG_OFF;
+    d->private->sg_hd=malloc(MAX_BIG_BUFF_SIZE);
+    d->private->sg_buffer=((unsigned char *)d->private->sg_hd)+SG_OFF;
   }
 
   {
@@ -747,14 +758,12 @@ cdrom_drive *cdda_identify_scsi(const char *generic_device,
   memcpy(d->inqbytes,p,4);
   d->cdda_device_name=copystring(generic_device);
   d->ioctl_device_name=copystring(specialized_device);
-
   d->drive_model=calloc(36,1);
   strscat(d->drive_model,p+8,8);
   strscat(d->drive_model,p+16,16);
   strscat(d->drive_model,p+32,4);
 
   idmessage(messagedest,messages,"\nCDROM model sensed sensed: %s",d->drive_model);
-  
   return(d);
   
 cdda_identify_scsi_fail:
@@ -763,7 +772,10 @@ cdda_identify_scsi_fail:
   if(i_fd!=-1)close(i_fd);
   if(g_fd!=-1)close(g_fd);
   if(d){
-    if(d->sg)free(d->sg);
+    if(d->private){
+      if(d->private->sg_hd)free(d->private->sg_hd);
+      free(d->private);
+    }
     free(d);
   }
   return(NULL);
@@ -809,6 +821,7 @@ cdrom_drive *cdda_identify_test(const char *filename, int messagedest,
   d->interface=TEST_INTERFACE;
   d->bigendianp=-1; /* We don't know yet... */
   d->nsectors=-1;
+  d->private=calloc(1,sizeof(*d->private));
   d->drive_model=copystring("File based test interface");
   idmessage(messagedest,messages,"\t\tCDROM sensed: %s\n",d->drive_model);
   
