@@ -1,5 +1,5 @@
 /******************************************************************
- * CopyPolicy: GNU Lesser General Public License 3 applies
+ * CopyPolicy: GNU Lesser General Public License 2.1 applies
  * Original interface.c Copyright (C) 1994-1997 
  *            Eissfeldt heiko@colossus.escape.de
  * Current incarnation Copyright (C) 1998-2008 Monty xiphmont@mit.edu
@@ -11,6 +11,20 @@
 #include "low_interface.h"
 #include "common_interface.h"
 #include "utils.h"
+#include <time.h>
+static int timed_ioctl(cdrom_drive *d, int fd, int command, void *arg){
+  struct timespec tv1;
+  struct timespec tv2;
+  int ret1=clock_gettime(d->private->clock,&tv1);
+  int ret2=ioctl(fd, command,arg);
+  int ret3=clock_gettime(d->private->clock,&tv2);
+  if(ret1<0 || ret3<0){
+    d->private->last_milliseconds=-1;
+  }else{
+    d->private->last_milliseconds = (tv2.tv_sec-tv1.tv_sec)*1000. + (tv2.tv_nsec-tv1.tv_nsec)/1000000.;
+  }
+  return ret2;
+}
 
 /* hook */
 static int Dummy (cdrom_drive *d,int s){
@@ -33,7 +47,7 @@ static void tweak_SG_buffer(cdrom_drive *d) {
   do {
     cur <<= 1; reserved = cur * (1<<9);
     err = ioctl(d->cdda_fd, SG_SET_RESERVED_SIZE, &reserved);
-  } while(err >= 0);
+  } while(err >= 0 && (cur*(1<<9) < 0x40000000));
   ioctl(d->cdda_fd, SG_GET_RESERVED_SIZE, &reserved);
 
   /* this doesn't currently ever work, but someday somebody might
@@ -42,31 +56,37 @@ static void tweak_SG_buffer(cdrom_drive *d) {
     table=1;
 
   sprintf(buffer,"\tDMA scatter/gather table entries: %d\n\t"
-      "table entry size: %d bytes\n\t"
-      "maximum theoretical transfer: %d sectors\n",
-      table, reserved, table*reserved/CD_FRAMESIZE_RAW);
+	  "table entry size: %d bytes\n\t"
+	  "maximum theoretical transfer: %d sectors\n",
+	  table, reserved, table*(reserved/CD_FRAMESIZE_RAW));
   cdmessage(d,buffer);
 
-  cur=table*reserved;
+  cur=reserved; /* only use one entry for now */
 
   /* so since we never go above q->max_sectors, we should never get -EIO.
    * we might still get -ENOMEM, but we back off for that later.  Monty
    * had an old comment: "not too much; new kernels have trouble with DMA
-   * "allocation, so be more conservative: 32kB max until I test more 
+   * allocation, so be more conservative: 32kB max until I test more 
    * thoroughly".  We're not currently honoring that, because we should
    * always get -ENOMEM.
    *
    * Updated: but we don't always get -ENOMEM.  Sometimes USB drives 
    * still fail the wrong way.  This needs some kernel-land investigation.
    */
+  /* Bumping to 64kB  transfer max --Monty */
+
   if (!getenv("CDDA_IGNORE_BUFSIZE_LIMIT")) {
-    cur=(cur>1024*32?1024*32:cur);
+    cur=(cur>1024*64?1024*64:cur);
+  }else{
+    cdmessage(d,"\tEnvironment variable CDDA_IGNORE_BUFSIZE_LIMIT set,\n"
+	      "\t\tforcing maximum possible sector size.  This can break\n"
+	      "\t\tspectacularly; use with caution!\n");
   }
   d->nsectors=cur/CD_FRAMESIZE_RAW;
   d->bigbuff=cur;
 
   sprintf(buffer,"\tSetting default read size to %d sectors (%d bytes).\n\n",
-      d->nsectors,d->nsectors*CD_FRAMESIZE_RAW);
+	  d->nsectors,d->nsectors*CD_FRAMESIZE_RAW);
 
   if(cur==0) exit(1);
 
@@ -76,7 +96,7 @@ static void tweak_SG_buffer(cdrom_drive *d) {
 static void clear_garbage(cdrom_drive *d){
   fd_set fdset;
   struct timeval tv;
-  struct sg_header *sg_hd=(struct sg_header *)d->sg;
+  struct sg_header *sg_hd=d->private->sg_hd;
   int flag=0;
 
   /* clear out any possibly preexisting garbage */
@@ -104,12 +124,16 @@ static void clear_garbage(cdrom_drive *d){
   }
 }
 
-static int check_sbp_error(const unsigned char *sbp) {
-  if (sbp[0]) {
-    char key = sbp[2] & 0xf;
-    char ASC = sbp[12];
-    char ASCQ = sbp[13];
-    
+static int check_sbp_error(const unsigned char status,
+			   const unsigned char *sbp) {
+  char key = sbp[2] & 0xf;
+  char ASC = sbp[12];
+  char ASCQ = sbp[13];
+  
+  if(status==0)return 0;
+  if(status==8)return TR_BUSY;
+
+  if (sbp[0]) {  
     switch (key){
     case 0:
       if (errno==0)
@@ -118,9 +142,8 @@ static int check_sbp_error(const unsigned char *sbp) {
     case 1:
       break;
     case 2:
-      if (errno==0)
-	errno = EBUSY;
-      return(TR_BUSY);
+      errno = ENOMEDIUM;
+      return(TR_NOTREADY);
     case 3: 
       if ((ASC==0x0C) & (ASCQ==0x09)) {
 	/* loss of streaming */
@@ -158,9 +181,11 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
 			       unsigned char bytefill,
 			       int bytecheck,
 			       unsigned char *sense_buffer){
-  
+  struct timespec tv1;
+  struct timespec tv2;
+  int tret1,tret2;
   int status = 0;
-  struct sg_header *sg_hd=(struct sg_header *)d->sg;
+  struct sg_header *sg_hd=d->private->sg_hd;
   long writebytes=SG_OFF+cmd_len+in_size;
 
   /* generic scsi device services */
@@ -170,7 +195,7 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
 
   memset(sg_hd,0,sizeof(sg_hd)); 
   memset(sense_buffer,0,SG_MAX_SENSE); 
-  memcpy(d->sg_buffer,cmd,cmd_len);
+  memcpy(d->private->sg_buffer,cmd,cmd_len+in_size);
   sg_hd->twelve_byte = cmd_len == 12;
   sg_hd->result = 0;
   sg_hd->reply_len = SG_OFF + out_size;
@@ -184,7 +209,7 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
      tell if the command failed.  Scared yet? */
 
   if(bytecheck && out_size>in_size){
-    memset(d->sg_buffer+cmd_len+in_size,bytefill,out_size-in_size); 
+    memset(d->private->sg_buffer+cmd_len+in_size,bytefill,out_size-in_size); 
     /* the size does not remove cmd_len due to the way the kernel
        driver copies buffers */
     writebytes+=(out_size-in_size);
@@ -192,8 +217,8 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
 
   {
     /* Select on write with a 5 second timeout.  This is a hack until
-       a better error reporting layer is in place in alpha 10; right
-       now, always print a message. */
+       a better error reporting layer is in place; right now, always
+       print a message. */
 
     fd_set fdset;
     struct timeval tv;
@@ -218,6 +243,7 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
   }
 
   sigprocmask (SIG_BLOCK, &(d->sigset), NULL );
+  tret1=clock_gettime(d->private->clock,&tv1);  
   errno=0;
   status = write(d->cdda_fd, sg_hd, writebytes );
 
@@ -226,11 +252,12 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
     if(errno==0)errno=EIO;
     return(TR_EWRITE);
   }
+
   
   {
     /* Select on read (and write; this signals an error) with a 5
        second timeout.  This is a hack until a better error reporting
-       layer is in place in alpha 10; right now, always print a
+       layer is in place; right now, always print a
        message. */
 
     fd_set rset;
@@ -262,19 +289,20 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
     }
   }
 
+  tret2=clock_gettime(d->private->clock,&tv2);  
   errno=0;
   status = read(d->cdda_fd, sg_hd, SG_OFF + out_size);
   sigprocmask ( SIG_UNBLOCK, &(d->sigset), NULL );
   memcpy(sense_buffer,sg_hd->sense_buffer,SG_MAX_SENSE);
 
-  if (status<0)return(TR_EREAD);
+  if (status<0)return status;
 
   if(status != SG_OFF + out_size || sg_hd->result){
     if(errno==0)errno=EIO;
     return(TR_EREAD);
   }
 
-  status = check_sbp_error(sense_buffer);
+  status = check_sbp_error(sg_hd->target_status, sense_buffer);
   if(status)return status;
 
   /* Failed/Partial DMA transfers occasionally get through.  Why?  No clue,
@@ -285,7 +313,7 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
   if(bytecheck && in_size+cmd_len<out_size){
     long i,flag=0;
     for(i=in_size;i<out_size;i++)
-      if(d->sg_buffer[i]!=bytefill){
+      if(d->private->sg_buffer[i]!=bytefill){
 	flag=1;
 	break;
       }
@@ -297,6 +325,11 @@ static int sg2_handle_scsi_cmd(cdrom_drive *d,
   }
 
   errno=0;
+  if(tret1<0 || tret2<0){
+    d->private->last_milliseconds=-1;
+  }else{
+    d->private->last_milliseconds = (tv2.tv_sec-tv1.tv_sec)*1000 + (tv2.tv_nsec-tv1.tv_nsec)/1000000;
+  }
   return(0);
 }
 
@@ -314,29 +347,31 @@ static int sgio_handle_scsi_cmd(cdrom_drive *d,
 
   memset(&hdr,0,sizeof(hdr));
   memset(sense,0,sizeof(sense));
-  
+  memcpy(d->private->sg_buffer,cmd+cmd_len,in_size);
+
   hdr.cmdp = cmd;
   hdr.cmd_len = cmd_len;
   hdr.sbp = sense;
   hdr.mx_sb_len = SG_MAX_SENSE;
   hdr.timeout = 50000;
   hdr.interface_id = 'S';
-  hdr.dxferp =  d->sg_buffer;
+  hdr.dxferp =  d->private->sg_buffer;
   hdr.flags = SG_FLAG_DIRECT_IO;  /* direct IO if we can get it */
 
   /* scary buffer fill hack */
   if(bytecheck && out_size>in_size)
     memset(hdr.dxferp+in_size,bytefill,out_size-in_size); 
 
-
   if (in_size) {
     hdr.dxfer_len = in_size;
     hdr.dxfer_direction = SG_DXFER_TO_DEV;
     
     errno = 0;
     status = ioctl(d->ioctl_fd, SG_IO, &hdr);
-    if (status >= 0 && hdr.status)
-      status = check_sbp_error(hdr.sbp);
+    if (status >= 0 && hdr.status){
+      status = check_sbp_error(hdr.status,hdr.sbp);
+      if(status) return status;
+    }
     if (status < 0) return TR_EWRITE;
   }
 
@@ -349,10 +384,12 @@ static int sgio_handle_scsi_cmd(cdrom_drive *d,
       hdr.dxfer_direction = out_size ? SG_DXFER_FROM_DEV : SG_DXFER_NONE;
 
     errno = 0;
-    status = ioctl(d->ioctl_fd, SG_IO, &hdr);
-    if (status >= 0 && hdr.status)
-      status = check_sbp_error(hdr.sbp);
-    if (status < 0) return TR_EREAD;
+    status = timed_ioctl(d,d->ioctl_fd, SG_IO, &hdr);
+    if (status >= 0 && hdr.status){
+      status = check_sbp_error(hdr.status,hdr.sbp);
+      if(status) return status;
+    }
+    if (status < 0) return status;
   }
 
   /* Failed/Partial DMA transfers occasionally get through.  Why?  No clue,
@@ -363,7 +400,7 @@ static int sgio_handle_scsi_cmd(cdrom_drive *d,
   if(bytecheck && in_size<out_size){
     long i,flag=0;
     for(i=in_size;i<out_size;i++)
-      if(d->sg_buffer[i]!=bytefill){
+      if(d->private->sg_buffer[i]!=bytefill){
 	flag=1;
 	break;
       }
@@ -373,7 +410,10 @@ static int sgio_handle_scsi_cmd(cdrom_drive *d,
       return(TR_ILLEGAL);
     }
   }
-  
+
+  /* Can't rely on .duration because we can't be certain kernel has HZ set to something useful */
+  /* d->private->last_milliseconds = hdr.duration; */
+
   errno = 0;
   return 0;
 }
@@ -405,9 +445,9 @@ static int test_unit_ready(cdrom_drive *d){
 
   handle_scsi_cmd(d, cmd, 6, 0, 56, 0,0, sense);
 
-  key = d->sg_buffer[2] & 0xf;
-  ASC = d->sg_buffer[12];
-  ASCQ = d->sg_buffer[13];
+  key = d->private->sg_buffer[2] & 0xf;
+  ASC = d->private->sg_buffer[12];
+  ASCQ = d->private->sg_buffer[13];
   
   if(key == 2 && ASC == 4 && ASCQ == 1) return 0;
   return 1;
@@ -452,7 +492,7 @@ static int mode_sense_atapi(cdrom_drive *d,int size,int page){
   if (handle_scsi_cmd (d, cmd, 10, 0, size+4,'\377',1,sense)) return(1);
 
   {
-    unsigned char *b=d->sg_buffer;
+    unsigned char *b=d->private->sg_buffer;
     if(b[0])return(1); /* Handles only up to 256 bytes */
     if(b[6])return(1); /* Handles only up to 256 bytes */
 
@@ -461,7 +501,7 @@ static int mode_sense_atapi(cdrom_drive *d,int size,int page){
     b[2]=b[3];
     b[3]=b[7];
 
-    memmove(b+4,b+8,size);
+    memmove(b+4,b+8,size-4);
   }
   return(0);
 }
@@ -482,6 +522,10 @@ static int mode_sense_scsi(cdrom_drive *d,int size,int page){
   cmd[4]=size;
   
   if (handle_scsi_cmd (d, cmd, 6, 0, size, '\377',1, sense)) return(1);
+
+  /* dump it all... */
+  
+
   return(0);
 }
 
@@ -491,7 +535,7 @@ static int mode_sense(cdrom_drive *d,int size,int page){
   return(mode_sense_scsi(d,size,page));
 }
 
-/* Current SG/SGIO impleemntations specifically disallow mode set
+/* Current SG/SGIO impleenmtations specifically disallow mode set
    unless running as root (or setuid).  One can see why (could be
    disastrous on, eg, a SCSI disk), but it curtails what we can do
    with older SCSI cdroms. */
@@ -507,7 +551,7 @@ static int mode_select(cdrom_drive *d,int density,int secsize){
 			      0,    /* reserved */
 			      0,    /* reserved */
 			      0,    /* reserved */
-			      12,   /* sizeof(mode) */
+			      16,   /* sizeof(mode) */
 			      0,    /* reserved */
 			      
 			      /* mode parameter header */
@@ -521,9 +565,6 @@ static int mode_select(cdrom_drive *d,int density,int secsize){
 			      0, 0, 0};/* Blocklen */
     unsigned char *mode = cmd + 18;
     
-    
-    cmd[1]|=d->lun<<5;
-    
     /* prepare to read cds in the previous mode */
     mode [0] = density;
     mode [6] = secsize >> 8;   /* block length "msb" */
@@ -563,8 +604,8 @@ static int mode_select(cdrom_drive *d,int density,int secsize){
 static unsigned int get_orig_sectorsize(cdrom_drive *d){
   if(mode_sense(d,12,0x01))return(-1);
 
-  d->orgdens = d->sg_buffer[4];
-  return(d->orgsize = ((int)(d->sg_buffer[10])<<8)+d->sg_buffer[11]);
+  d->orgdens = d->private->sg_buffer[4];
+  return(d->orgsize = ((int)(d->private->sg_buffer[10])<<8)+d->private->sg_buffer[11]);
 }
 
 /* switch CDROM scsi drives to given sector size  */
@@ -572,7 +613,7 @@ static int set_sectorsize (cdrom_drive *d,unsigned int secsize){
   return(mode_select(d,d->orgdens,secsize));
 }
 
-/* switch Toshiba/DEC and HP drives from/to cdda density */
+/* switch old Toshiba/DEC and HP drives from/to cdda density */
 int scsi_enable_cdda (cdrom_drive *d, int fAudioMode){
   if (fAudioMode) {
     if(mode_select(d,d->density,CD_FRAMESIZE_RAW)){
@@ -623,8 +664,8 @@ static int scsi_read_toc (cdrom_drive *d){
     return(-4);
   }
 
-  first=d->sg_buffer[2];
-  last=d->sg_buffer[3];
+  first=d->private->sg_buffer[2];
+  last=d->private->sg_buffer[3];
   tracks=last-first+1;
 
   if (last > MAXTRK || first > MAXTRK || last<0 || first<0) {
@@ -642,7 +683,7 @@ static int scsi_read_toc (cdrom_drive *d){
       return(-5);
     }
     {
-      scsi_TOC *toc=(scsi_TOC *)(d->sg_buffer+4);
+      scsi_TOC *toc=(scsi_TOC *)(d->private->sg_buffer+4);
 
       d->disc_toc[i-first].bFlags=toc->bFlags;
       d->disc_toc[i-first].bTrack=i;
@@ -663,7 +704,7 @@ static int scsi_read_toc (cdrom_drive *d){
     return(-2);
   }
   {
-    scsi_TOC *toc=(scsi_TOC *)(d->sg_buffer+4);
+    scsi_TOC *toc=(scsi_TOC *)(d->private->sg_buffer+4);
     
     d->disc_toc[i-first].bFlags=toc->bFlags;
     d->disc_toc[i-first].bTrack=0xAA;
@@ -697,7 +738,7 @@ static int scsi_read_toc2 (cdrom_drive *d){
   }
 
   /* copy to our structure and convert start sector */
-  tracks = d->sg_buffer[1];
+  tracks = d->private->sg_buffer[1];
   if (tracks > MAXTRK) {
     cderror(d,"003: CDROM reporting illegal number of tracks\n");
     return(-3);
@@ -713,39 +754,52 @@ static int scsi_read_toc2 (cdrom_drive *d){
       return(-5);
     }
     
-    d->disc_toc[i].bFlags = d->sg_buffer[10];
+    d->disc_toc[i].bFlags = d->private->sg_buffer[10];
     d->disc_toc[i].bTrack = i + 1;
 
     d->disc_toc[i].dwStartSector= d->adjust_ssize * 
-	(((signed char)(d->sg_buffer[2])<<24) | 
-	 (d->sg_buffer[3]<<16)|
-	 (d->sg_buffer[4]<<8)|
-	 (d->sg_buffer[5]));
+	(((signed char)(d->private->sg_buffer[2])<<24) | 
+	 (d->private->sg_buffer[3]<<16)|
+	 (d->private->sg_buffer[4]<<8)|
+	 (d->private->sg_buffer[5]));
   }
 
   d->disc_toc[i].bFlags = 0;
   d->disc_toc[i].bTrack = i + 1;
-  memcpy (&foo, d->sg_buffer+2, 4);
-  memcpy (&bar, d->sg_buffer+6, 4);
+  memcpy (&foo, d->private->sg_buffer+2, 4);
+  memcpy (&bar, d->private->sg_buffer+6, 4);
   d->disc_toc[i].dwStartSector = d->adjust_ssize * (be32_to_cpu(foo) +
 						    be32_to_cpu(bar));
 
   d->disc_toc[i].dwStartSector= d->adjust_ssize * 
-    ((((signed char)(d->sg_buffer[2])<<24) | 
-      (d->sg_buffer[3]<<16)|
-      (d->sg_buffer[4]<<8)|
-      (d->sg_buffer[5]))+
+    ((((signed char)(d->private->sg_buffer[2])<<24) | 
+      (d->private->sg_buffer[3]<<16)|
+      (d->private->sg_buffer[4]<<8)|
+      (d->private->sg_buffer[5]))+
      
-     ((((signed char)(d->sg_buffer[6])<<24) | 
-       (d->sg_buffer[7]<<16)|
-       (d->sg_buffer[8]<<8)|
-       (d->sg_buffer[9]))));
+     ((((signed char)(d->private->sg_buffer[6])<<24) | 
+       (d->private->sg_buffer[7]<<16)|
+       (d->private->sg_buffer[8]<<8)|
+       (d->private->sg_buffer[9]))));
 
 
   d->cd_extra = FixupTOC(d,tracks+1);
   return(tracks);
 }
 
+static int scsi_set_speed (cdrom_drive *d, int speed){
+  unsigned char cmd[12]={0xBB, 0, 0, 0, 0xff, 0xff, 0, 0, 0, 0, 0, 0};
+  unsigned char sense[SG_MAX_SENSE];
+
+  if(speed>=0)
+    speed=speed*44100*4/1024;
+  else
+    speed=-1;
+  cmd[2] = (speed >> 8) & 0xFF;
+  cmd[3] = (speed) & 0xFF;
+  return handle_scsi_cmd(d,cmd,12,0,0,0,0,sense);
+}
+
 /* These do one 'extra' copy in the name of clean code */
 
 static int i_read_28 (cdrom_drive *d, void *p, long begin, long sectors, unsigned char *sense){
@@ -763,7 +817,7 @@ static int i_read_28 (cdrom_drive *d, void *p, long begin, long sectors, unsigne
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,10,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -782,7 +836,7 @@ static int i_read_A8 (cdrom_drive *d, void *p, long begin, long sectors, unsigne
   cmd[9] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -800,7 +854,7 @@ static int i_read_D4_10 (cdrom_drive *d, void *p, long begin, long sectors, unsi
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,10,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -818,7 +872,7 @@ static int i_read_D4_12 (cdrom_drive *d, void *p, long begin, long sectors, unsi
   cmd[9] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -836,7 +890,7 @@ static int i_read_D5 (cdrom_drive *d, void *p, long begin, long sectors, unsigne
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,10,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -854,7 +908,7 @@ static int i_read_D8 (cdrom_drive *d, void *p, long begin, long sectors, unsigne
   cmd[9] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -868,7 +922,7 @@ static int i_read_mmc (cdrom_drive *d, void *p, long begin, long sectors, unsign
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -882,7 +936,7 @@ static int i_read_mmcB (cdrom_drive *d, void *p, long begin, long sectors, unsig
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -896,7 +950,7 @@ static int i_read_mmc2 (cdrom_drive *d, void *p, long begin, long sectors, unsig
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -910,7 +964,7 @@ static int i_read_mmc2B (cdrom_drive *d, void *p, long begin, long sectors, unsi
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -924,7 +978,7 @@ static int i_read_mmc3 (cdrom_drive *d, void *p, long begin, long sectors, unsig
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -938,7 +992,7 @@ static int i_read_mmc3B (cdrom_drive *d, void *p, long begin, long sectors, unsi
   cmd[8] = sectors;
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -972,7 +1026,7 @@ static int i_read_msf (cdrom_drive *d, void *p, long begin, long sectors, unsign
 
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -985,7 +1039,7 @@ static int i_read_msf2 (cdrom_drive *d, void *p, long begin, long sectors, unsig
 
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -998,7 +1052,7 @@ static int i_read_msf3 (cdrom_drive *d, void *p, long begin, long sectors, unsig
   
   if((ret=handle_scsi_cmd(d,cmd,12,0,sectors * CD_FRAMESIZE_RAW,'\177',1,sense)))
     return(ret);
-  if(p)memcpy(p,d->sg_buffer,sectors*CD_FRAMESIZE_RAW);
+  if(p)memcpy(p,d->private->sg_buffer,sectors*CD_FRAMESIZE_RAW);
   return(0);
 }
 
@@ -1017,6 +1071,7 @@ static long scsi_read_map (cdrom_drive *d, void *p, long begin, long sectors,
   retry_count=0;
   
   while(1) {
+
     if((err=map(d,(p?buffer:NULL),begin,sectors,sense))){
       if(d->report_all){
 	char b[256];
@@ -1043,8 +1098,7 @@ static long scsi_read_map (cdrom_drive *d, void *p, long begin, long sectors,
 	fprintf(stderr,"                 Transport error: %s\n",strerror_tr[err]);
 	fprintf(stderr,"                 System error: %s\n",strerror(errno));
       }
-
-      if(!d->error_retry)return(-7);
+      
       switch(errno){
       case EINTR:
 	usleep(100);
@@ -1066,6 +1120,10 @@ static long scsi_read_map (cdrom_drive *d, void *p, long begin, long sectors,
 	}
 	sectors--;
 	continue;
+      case ENOMEDIUM:
+	cderror(d,"404: No medium present\n");
+	return(-404);
+
       default:
 	if(sectors==1){
 	  if(errno==EIO)
@@ -1091,6 +1149,8 @@ static long scsi_read_map (cdrom_drive *d, void *p, long begin, long sectors,
            back to cdda */
 	reset_scsi(d);
       }
+      if(!d->error_retry)return(-7);
+
     }else{
 
       /* Did we get all the bytes we think we did, or did the kernel
@@ -1114,6 +1174,7 @@ static long scsi_read_map (cdrom_drive *d, void *p, long begin, long sectors,
 	}
 	
 	if(i>0)return(i);
+       
       }else
 	break;
     }
@@ -1214,7 +1275,7 @@ long scsi_read_msf3 (cdrom_drive *d, void *p, long begin,
 static int count_2352_bytes(cdrom_drive *d){
   long i;
   for(i=2351;i>=0;i--)
-    if(d->sg_buffer[i]!=(unsigned char)'\177')
+    if(d->private->sg_buffer[i]!=(unsigned char)'\177')
       return(((i+3)>>2)<<2);
 
   return(0);
@@ -1223,7 +1284,7 @@ static int count_2352_bytes(cdrom_drive *d){
 static int verify_nonzero(cdrom_drive *d){
   long i,flag=0;
   for(i=0;i<2352;i++)
-    if(d->sg_buffer[i]!=0){
+    if(d->private->sg_buffer[i]!=0){
       flag=1;
       break;
     }
@@ -1273,13 +1334,11 @@ static int verify_read_command(cdrom_drive *d){
     
     d->enable_cdda(d,0);
   }
-
   if(!audioflag){
     cdmessage(d,"\tCould not find any audio tracks on this disk.\n");
     return(-403);
   }
 
-
   {
     char *es="",*rs="";
     d->bigendianp=-1;
@@ -1375,7 +1434,6 @@ static int verify_read_command(cdrom_drive *d){
       case 14:
 	d->read_audio=scsi_read_D8;
 	rs="d8 0x,00";
-	j=-2;
 	break;
       }
       
@@ -1385,7 +1443,7 @@ static int verify_read_command(cdrom_drive *d){
 	  d->density=0;
 	  d->enable_cdda=Dummy;
 	  es="none    ";
-	  if(!densitypossible)i=-2; /* short circuit MMC style commands */
+	  if(!densitypossible)i=5; /* short circuit MMC style commands */
 	  break;
 	case 1:
 	  d->density=0;
@@ -1401,11 +1459,11 @@ static int verify_read_command(cdrom_drive *d){
 	  d->density=0x82;
 	  d->enable_cdda=scsi_enable_cdda;
 	  es="yes/0x82";
+	  break;
 	case 4:
 	  d->density=0x81;
 	  d->enable_cdda=scsi_enable_cdda;
 	  es="yes/0x81";
-	  i=-2;
 	  break;
 	}
 
@@ -1489,42 +1547,44 @@ static int verify_read_command(cdrom_drive *d){
   return(-6);
 }
 
-static void check_fua_bit(cdrom_drive *d){
-  int16_t *buff=malloc(CD_FRAMESIZE_RAW);
+static void check_cache(cdrom_drive *d){
   long i;
 
-  if(d->read_audio==scsi_read_mmc)return;
-  if(d->read_audio==scsi_read_mmc2)return;
-  if(d->read_audio==scsi_read_mmc3)return;
-  if(d->read_audio==scsi_read_mmcB)return;
-  if(d->read_audio==scsi_read_mmc2B)return;
-  if(d->read_audio==scsi_read_mmc3B)return;
+  if(!(d->read_audio==scsi_read_mmc ||
+     d->read_audio==scsi_read_mmc2 ||
+     d->read_audio==scsi_read_mmc3 ||
+     d->read_audio==scsi_read_mmcB ||
+     d->read_audio==scsi_read_mmc2B ||
+       d->read_audio==scsi_read_mmc3B)){
 
-  cdmessage(d,"This command set may use a Force Unit Access bit.");
-  cdmessage(d,"\nChecking drive for FUA bit support...\n");
-  
-  d->enable_cdda(d,1);
-  d->fua=1;
-  
-  for(i=1;i<=d->tracks;i++){
-    if(cdda_track_audiop(d,i)==1){
-      long firstsector=cdda_track_firstsector(d,i);
-      long lastsector=cdda_track_lastsector(d,i);
-      long sector=(firstsector+lastsector)>>1;
-      
-      if(d->read_audio(d,buff,sector,1)>0){
-	cdmessage(d,"\tDrive accepted FUA bit.\n");
-	d->enable_cdda(d,0);
-	free(buff);
-	return;
+    cdmessage(d,"This command set may use a Force Unit Access bit.");
+    cdmessage(d,"\nChecking drive for FUA bit support...\n");
+    
+    d->enable_cdda(d,1);
+    d->fua=1;
+    
+    for(i=1;i<=d->tracks;i++){
+      if(cdda_track_audiop(d,i)==1){
+	long firstsector=cdda_track_firstsector(d,i);
+	long lastsector=cdda_track_lastsector(d,i);
+	long sector=(firstsector+lastsector)>>1;
+	
+	if(d->read_audio(d,NULL,sector,1)>0){
+	  cdmessage(d,"\tDrive accepted FUA bit.\n");
+	  d->enable_cdda(d,0);
+	  return;
+	}
       }
     }
+    
+    d->fua=0;
+    cdmessage(d,"\tDrive rejected FUA bit.\n");
+
+    /* we only use the FUA bit as a possible extra layer of
+       redundancy; too many drives accept it, but still don't force
+       unit access. Still use the old cachebusting algo. */
+    return;
   }
-  
-  d->fua=0;
-  cdmessage(d,"\tDrive rejected FUA bit.\n");
-  free(buff);
-  return;
 }
 
 static int check_atapi(cdrom_drive *d){
@@ -1565,7 +1625,7 @@ static int check_mmc(cdrom_drive *d){
   d->is_mmc=0;
   if(mode_sense(d,22,0x2A)==0){
   
-    b=d->sg_buffer;
+    b=d->private->sg_buffer;
     b+=b[3]+4;
     
     if((b[0]&0x3F)==0x2A){
@@ -1613,10 +1673,9 @@ unsigned char *scsi_inquiry(cdrom_drive *d){
     cderror(d,"008: Unable to identify CDROM model\n");
     return(NULL);
   }
-  return (d->sg_buffer);
+  return (d->private->sg_buffer);
 }
 
-
 int scsi_init_drive(cdrom_drive *d){
   int ret;
 
@@ -1660,8 +1719,7 @@ int scsi_init_drive(cdrom_drive *d){
 
   d->read_toc = (!memcmp(d->drive_model, "IMS", 3) && !d->is_atapi) ? scsi_read_toc2 : 
     scsi_read_toc;
-  d->set_speed = NULL;
-  
+  d->set_speed = scsi_set_speed;
 
   if(!d->is_atapi){
     unsigned sector_size= get_orig_sectorsize(d);
@@ -1681,11 +1739,11 @@ int scsi_init_drive(cdrom_drive *d){
   d->opened=1;
 
   if((ret=verify_read_command(d)))return(ret);
-  check_fua_bit(d);
+  check_cache(d);
 
   d->error_retry=1;
-  d->sg=realloc(d->sg,d->nsectors*CD_FRAMESIZE_RAW + SG_OFF + 128);
-  d->sg_buffer=d->sg+SG_OFF;
+  d->private->sg_hd=realloc(d->private->sg_hd,d->nsectors*CD_FRAMESIZE_RAW + SG_OFF + 128);
+  d->private->sg_buffer=((unsigned char *)d->private->sg_hd)+SG_OFF;
   d->report_all=1;
   return(0);
 }
