@@ -1,5 +1,5 @@
 /***
- * CopyPolicy: GNU Lesser General Public License 3 applies
+ * CopyPolicy: GNU Lesser General Public License 2.1 applies
  * Copyright (C) by Monty (xiphmont@mit.edu)
  *
  * Toplevel file for the paranoia abstraction over the cdda lib 
@@ -76,11 +76,15 @@
 #include <math.h>
 #include "../interface/cdda_interface.h"
 #include "../interface/smallft.h"
+#include "../version.h"
 #include "p_block.h"
 #include "cdda_paranoia.h"
 #include "overlap.h"
 #include "gap.h"
 #include "isort.h"
+#include <errno.h>
+
+#define MIN_SEEK_MS 6
 
 static inline long re(root_block *root){
   if(!root)return(-1);
@@ -125,7 +129,6 @@ enum  {
   FLAGS_VERIFIED=0x4  /**< block read and verified */
 } paranoia_read_flags;
 
-
 /**** matching and analysis code *****************************************/
 
 /* ===========================================================================
@@ -183,7 +186,8 @@ static inline long i_paranoia_overlap(int16_t *buffA,int16_t *buffB,
  * offsets of the first and last matching samples in A.
  */
 static inline long i_paranoia_overlap2(int16_t *buffA,int16_t *buffB,
-				       char *flagsA,char *flagsB,
+				       unsigned char *flagsA,
+				       unsigned char *flagsB,
 				       long offsetA, long offsetB,
 				       long sizeA,long sizeB,
 				       long *ret_begin, long *ret_end){
@@ -252,10 +256,11 @@ static inline long i_paranoia_overlap2(int16_t *buffA,int16_t *buffB,
  * on the CD and what B considers sample N.)
  */
 static inline long do_const_sync(c_block *A,
-				 sort_info *B,char *flagB,
+				 sort_info *B,
+				 unsigned char *flagB,
 				 long posA,long posB,
 				 long *begin,long *end,long *offset){
-  char *flagA=A->flags;
+  unsigned char *flagA=A->flags;
   long ret=0;
 
   /* If we're doing any verification whatsoever, we have flags in stage
@@ -318,14 +323,14 @@ static inline long do_const_sync(c_block *A,
    reference */
 
 static inline long try_sort_sync(cdrom_paranoia *p,
-				 sort_info *A,char *Aflags,
+				 sort_info *A,unsigned char *Aflags,
 				 c_block *B,
 				 long post,long *begin,long *end,
 				 long *offset,void (*callback)(long,int)){
   
   long dynoverlap=p->dynoverlap;
   sort_link *ptr=NULL;
-  char *Bflags=B->flags;
+  unsigned char *Bflags=B->flags;
 
   /* block flag matches FLAGS_UNREAD (and hence unmatchable) */
   if(Bflags==NULL || (Bflags[post-cb(B)]&FLAGS_UNREAD)==0){
@@ -2222,6 +2227,54 @@ long paranoia_seek(cdrom_paranoia *p,long seek,int mode){
   return(ret);
 }
 
+static void cdrom_cache_update(cdrom_paranoia *p, int lba, int sectors){
+
+  if(lba+sectors > p->cdcache_size){
+    int end = lba+sectors;
+    lba=end-p->cdcache_size;
+    sectors = end-lba;
+  }
+    
+  if(lba < p->cdcache_begin){
+    /* a backseek flushes the cache */
+    p->cdcache_begin=lba;
+    p->cdcache_end=lba+sectors;
+  }else{
+    if(lba+sectors>p->cdcache_end)
+      p->cdcache_end = lba+sectors;
+    if(lba+sectors-p->cdcache_size > p->cdcache_begin){
+      if(lba+sectors-p->cdcache_size < p->cdcache_end){
+	p->cdcache_begin = lba+sectors-p->cdcache_size;
+      }else{
+	p->cdcache_begin = lba;
+      }
+    }
+  }
+}
+
+static void cdrom_cache_handler(cdrom_paranoia *p, int lba, void(*callback)(long,int)){
+  int seekpos;
+  int ms;
+  if(lba>=p->cdcache_end)return; /* nothing to do */
+
+  if(lba<0)lba=0;
+
+  if(lba<p->cdcache_begin){
+    /* should always trigger a backseek so let's do that here and look for the timing */
+    seekpos=(lba==0 || lba-1<cdda_disc_firstsector(p->d) ? lba : lba-1); /* keep reads linear when possible */
+  }else{
+    int pre = p->cdcache_begin-1;
+    int post = lba+p->cdcache_size;
+
+    seekpos = (pre<cdda_disc_firstsector(p->d) ? post : pre);
+  }
+
+  if(cdda_read_timed(p->d,NULL,seekpos,1,&ms)==1)
+    if(seekpos<p->cdcache_begin && ms<MIN_SEEK_MS)
+      callback(seekpos*CD_FRAMEWORDS,PARANOIA_CB_CACHEERR);
+  cdrom_cache_update(p,seekpos,1);
+  return;
+}
 
 
 /* ===========================================================================
@@ -2266,13 +2319,13 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
    drives with unaddressable sectors behave more often). */
       
   long readat,firstread;
-  long totaltoread=p->readahead;
+  long totaltoread=p->cdcache_size;
   long sectatonce=p->d->nsectors;
   long driftcomp=(float)p->dyndrift/CD_FRAMEWORDS+.5;
   c_block *new=NULL;
   root_block *root=&p->root;
   int16_t *buffer=NULL;
-  char *flags=NULL;
+  unsigned char *flags=NULL;
   long sofar;
   long dynoverlap=(p->dynoverlap+CD_FRAMEWORDS-1)/CD_FRAMEWORDS; 
   long anyflag=0;
@@ -2289,16 +2342,12 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
   
   if(p->enable&(PARANOIA_MODE_VERIFY|PARANOIA_MODE_OVERLAP)){
     
-    /* we want to jitter the read alignment boundary */
     long target;
     if(rv(root)==NULL || rb(root)>beginword)
       target=p->cursor-dynoverlap; 
     else
       target=re(root)/(CD_FRAMEWORDS)-dynoverlap;
 	
-    if(target+MIN_SECTOR_BACKUP>p->lastread && target<=p->lastread)
-      target=p->lastread-MIN_SECTOR_BACKUP;
-      
     /* we want to jitter the read alignment boundary, as some
        drives, beginning from a specific point, will tend to
        lose bytes between sectors in the same place.  Also, as
@@ -2307,8 +2356,8 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
     
     readat=(target&(~((long)JIGGLE_MODULO-1)))+p->jitter;
     if(readat>target)readat-=JIGGLE_MODULO;
-    p->jitter++;
-    if(p->jitter>=JIGGLE_MODULO)p->jitter=0;
+    p->jitter--;
+    if(p->jitter<0)p->jitter+=JIGGLE_MODULO;
      
   }else{
     readat=p->cursor; 
@@ -2334,19 +2383,15 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
   sofar=0;
   firstread=-1;
   
-  /* Issue each of the low-level reads until we've read enough sectors
-   * to exhaust the drive's cache.
+  /* we have a read span; flush the drive cache if needed */
+  cdrom_cache_handler(p, readat, callback);
+
+  /* Issue each of the low-level reads; the optimal read size is
+   * approximately the cachemodel's cdrom cache size.  The only reason
+   * to read less would be memory considerations.
    *
    * p->readahead   = total number of sectors to read
    * p->d->nsectors = number of sectors to read per request
-   *
-   * The driver determines this latter number, which is the maximum
-   * number of sectors the kernel can reliably read per request.  In
-   * old Linux kernels, there was a hard limit of 8 sectors per read.
-   * While this limit has since been removed, certain motherboards
-   * can't handle DMA requests larger than 64K.  And other operating
-   * systems may have similar limitations.  So the method of splitting
-   * up reads is still useful.
    */
 
   /* actual read loop */
@@ -2389,10 +2434,20 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
        * cases, you take what part of the read you know is good, and
        * you get substantially better performance. --Monty
        */
+
       if((thisread=cdda_read(p->d,buffer+sofar*CD_FRAMEWORDS,adjread,
 			    secread))<secread){
 
-	if(thisread<0)thisread=0;
+	if(thisread<0){
+	  if(errno==ENOMEDIUM){
+	    /* the one error we bail on immediately */
+	    if(new)free_c_block(new);
+	    if(buffer)free(buffer);
+	    if(flags)free(flags);
+	    return NULL;
+	  }
+	  thisread=0;
+	}
 
 	/* Uhhh... right.  Make something up. But don't make us seek
            backward! */
@@ -2427,21 +2482,12 @@ c_block *i_read_c_block(cdrom_paranoia *p,long beginword,long endword,
 	  flags[sofar*CD_FRAMEWORDS+i]|=FLAGS_EDGE;
       }
 
-
-      /* Move the read cursor ahead by the number of sectors we attempted
-       * to read.
-       *
-       * "???: Again, why not move it ahead by the number actually
-       * read?"  Because adding zero would not be moving
-       * ahead. --Monty
-       */
-      p->lastread=adjread+secread;
-      
       if(adjread+secread-1==p->current_lastsector)
 	new->lastsector=-1;
       
       if(callback)(*callback)((adjread+secread-1)*CD_FRAMEWORDS,PARANOIA_CB_READ);
       
+      cdrom_cache_update(p,adjread,secread);
       sofar+=secread;
       readat=adjread+secread; 
     }else /* secread <= 0 */
@@ -2502,6 +2548,11 @@ int16_t *paranoia_read_limited(cdrom_paranoia *p, void(*callback)(long,int),
   long retry_count=0,lastend=-2;
   root_block *root=&p->root;
 
+  if(p->d->opened==0){
+    errno=EBADF;
+    return NULL;
+  }
+
   if(beginword>p->root.returnedlimit)p->root.returnedlimit=beginword;
   lastend=re(root);
 
@@ -2632,6 +2683,12 @@ int16_t *paranoia_read_limited(cdrom_paranoia *p, void(*callback)(long,int),
 			  callback);
       
 	}
+      }else{
+
+	/* Was the medium removed or the device closed out from
+	   under us? */
+	if(errno==ENOMEDIUM) return NULL;
+      
       }
     }
 
@@ -2687,3 +2744,7 @@ void paranoia_overlapset(cdrom_paranoia *p, long overlap){
   p->dynoverlap=overlap*CD_FRAMEWORDS;
   p->stage1.offpoints=-1; 
 }
+
+char *paranoia_version(void){
+  return VERSIONNUM;
+}
